# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off

import functools
import pathlib

@functools.lru_cache(maxsize=None)
def _load_schema(schema_name: str) -> str:
    return (pathlib.Path(__file__).parent / f"{schema_name}.avsc").read_text()

def getMetadataChangeEventSchema() -> str:
    return _load_schema("MetadataChangeEvent")

def getMetadataChangeProposalSchema() -> str:
    return _load_schema("MetadataChangeProposal")

# fmt: on
